% ---------------------------------------------------------------------------------------
%
% function [ MV[, INIT ] = LIFE_GAME( [BSZ[, PER[, EPOCHS[, TM[, FPS]]]]])
%
%          Executes the algorithm for the game of life as proposed by John Conway (1970)
%          and produces a 'movie' MV to be seen with the MATLAB function movie.
%
%          The function constructs a square binary image of size BSZ with an initial
%          percentage of true pixels (objects) controlled by the percentage variable PER,
%          then runs the game-of-life algorithm until no more changes are detected or
%          the number of repetitions reaches a maximun provided by the variable EPOCHS.
%          The inital image is produced using a variable drawn by a normal distribution.
%          Variables TM and FPS control the repetitions and framerate of the final movie
%          presentation in case no output arguments are defined.
%          The function also returns the initial binary image INIT for reference to 
%          the initial conditions of the game.
%
%          Boundary conditions: before each iteration the image canvas is augmented by 2 pixels row
%          and column-wise and the image is padded with zero values
%
%          Default variable values are:
%                 BSZ=100, PER=5, EPOCHS=200, TM=1, FPS=25
%
%          Example:
%                 M = life_game( 100, 12, 100);
%                 M = life_game( 100, 17, 1000);
%                 life_game( 100, 10, 100, 1, 25);
%
%          WARNING: the code is not optimized at all and the BSZ size has a huge impact on the
%          execution times, so it has to be kept to relatively small values.
%

%
% Copyright (C) 16-Feb-2014 George P. Pavlidis(R) (info@georgepavlides.info, gpavlid@gmail.com)
%
function varargout = life_game( varargin)

switch nargin
	case 0
		sz = 100;
		percentage = 5;
		epochs = 200;
		times = 1;
		fps = 25;
	case 1
		sz = varargin{1};
		percentage = 5;
		epochs = 200;
		times = 1;
		fps = 25;
	case 2
		sz = varargin{1};
		percentage = varargin{2};
		epochs = 200;
		times = 1;
		fps = 25;
	case 3
		sz = varargin{1};
		percentage = varargin{2};
		epochs = varargin{3};
		times = 1;
		fps = 25;
	case 4
		sz = varargin{1};
		percentage = varargin{2};
		epochs = varargin{3};
		times = varargin{4};
	otherwise
		sz = varargin{1};
		percentage = varargin{2};
		epochs = varargin{3};
		times = varargin{4};
		fps = varargin{5};
end
if percentage > 100
	error('Percentage cannot be greater than 100%!');
end
SS = get(0,'ScreenSize');
Height = SS(4);

a = randi([-round(sz/(percentage)) -1],sz)>-2;
init = a;

figure(2);
set( 2, 'Position', [660 Height-400-100 450 400]);
colormap( gray);
figure(1);
set( 1, 'Position', [50 Height-400-100 600 400]);
colormap( gray);
subplot(121);
imagesc( init);
axis image;
title( 'Initial conditions');

converged = false;
e = 1;
c = zeros( sz);
while ~converged && e<=epochs
	b = zeros( sz);
	a = pad_image( a, 2, 2); 
	for i=2:sz+1,
		for j=2:sz+1,
			w=a(i-1:i+1,j-1:j+1);
			s=sum(w(:))-a(i,j);
			% this is the core of the life game algorithm
			% a pixel survives if it was alive and has two alive neighbours 
			% or if it is any pixel (either dead or alive) with three alive neighbours
			if (a(i,j) && (s==2)) || (s==3)
				b(i-1,j-1)=1;
			end
		end
	end
	a = a(2:end-1,2:end-1);
	if a==b
		disp( 'there cannot be any further progress for life in this universe...');
		converged = true;
	end
	if b==c
		disp( 'life in this universe is looping between two conditions...');
		converged = true;
	end
	if sum(b(:))==0
		disp( 'all life died out in this universe...');
		converged = true;
	end
	c = a;
	a = b;
	if e==1
		figure(1);
		subplot(122);
		imagesc( a);
		axis image;
		title( 'First epoch conditions');
	end
	figure(2);
	imagesc( a);
	axis image;
	title( sprintf( 'Epoch #%d', e));
	M(e) = getframe;
	e = e+1;
end

switch nargout
	case 0
		disp('done.');
		disp('running the movie');
		figure(3);
		set( 3, 'Position', [660 Height-400-100 450 400]);
		axis off;
		movie(M,times,fps);
		disp('movie finished playback.');
	case 1
		varargout{1} = M;
	otherwise
		varargout{1} = M;
		varargout{2} = init;
end


% ---------------------------------------------------------------------------------------
%
%  function OUT_IM = PAD_IMAGE( IN_IM, ROW_OFFSET, COL_OFFSET)
%
%           auxiliary function to zero-pad images
%           IN_IM is the input image
%           ROW_OFFSET and COL_OFFSET are the amount of pixels to enlarge the image canvas
%
function img = pad_image( img, row_ofs, col_ofs)
	im = zeros( size( img,1)+row_ofs, size(img,2)+col_ofs);
	im(2:end-1,2:end-1) = img;
	img = im;

	
% ---------------------------------------------------------------------------------------
%
%  optimization notes:
%      should focus on the active areas... 
%      could use find to get the white pixels and work on 9x9 areas around these pixels
%      but what about the black pixels near the white ones?
%
%f=find(a);
%w=[i-rows-1:i-rows+1;i-1:i+1;i+rows-1:i+rows+1]'